#!/usr/bin/env python
#
# Data Quality Checker also known as DQC
#
# Read the README.md for more detailed information
#
#       Barcelona Supercomputing Center - Centro Nacional de Supercomputacion
#       Earth Sciences Department - Computational Earth Sciences
#     
#     Authors: 
#       Joan Sala Calero (joan.sala@bsc.es)
#       Jesus Pena-Izquierdo (jesus.pena@bsc.es)

import os
import pandas as pd
import numpy as np

from bokeh.plotting import figure
from bokeh.layouts import gridplot, column
from bokeh.io import show, save, output_file
from bokeh.models import ColumnDataSource, HoverTool, Span

from utils.utils_files import *

"""
Class: Bokeh plots
---
- General plot functions for histograms 
- Statistical formulas needed to produce the plots
"""
class bokehPlot:

    def __init__(self, outfname, csvs, csvh, xl, vmin, vmax):
        self.outfile_hist = outfname
        self.csv_hist = csvh
        self.csv_stat = csvs
        self.xaxis_label = xl
        self.valid_min = vmin
        self.valid_max = vmax
        self.regions = {
          'total' : '', 
          'N_polar' : '[60 to 90 deg]', 
          'N_midlat' : '[30 to 60 deg]', 
          'N_hadley' : '[0 to 30 deg]', 
          'S_hadley' : '[0 to -30 deg]', 
          'S_midlat' : '[-30 to -60 deg]', 
          'S_polar' : '[-60 to -90 deg]'
        }

    """
    Adjusted boxplot
    """
    def adj_boxplot(self,q1,q3,skew, dev, Ks, Ke):
        IQD = q3 - q1
        if IQD == 0:
            IQD = dev
        return q1 - Ke*IQD*np.exp(-Ks*skew1), q3 + Ke*IQD*np.exp(Ks*skew1)

    def std_boxplot(self,m, dev, K):
        return m -dev*K, m + dev*K

    def iqd_boxplot(self,q1,q3,K = 1.5):
        IQD = q3 - q1
        if IQD == 0:
            IQD = dev
        return q1 - K*IQD, q3 + K*IQD

    def hist2array(self,xm, n):
        x = [[ xi]*int(ni) for (xi, ni) in zip(xm, n)]
        return np.hstack(x)

    def hist2array_cut(self,xm, n, rcut = 0.001):
        Nt = np.sum(n)
        ncum = np.cumsum(n/Nt)
        i_nr0i = np.argwhere( ncum <= rcut ).flatten()
        i_nr0d = np.argwhere( ncum >= (1-rcut)).flatten()
        i_nr0 = np.concatenate((i_nr0i, i_nr0d))
        i_nr1 = np.argwhere((ncum > rcut) & (ncum < 1-rcut)).flatten()
        n_r0 = n[i_nr0]
        xm_r0 = xm[i_nr0]
        max_r0id = [n[i_nr0i[-1]], n[i_nr0d[0]]]
        if 0 in max_r0id:
            max_r0id.remove(0)
        k_r0 = np.min(max_r0id)
        print(k_r0)
        n_r1 = np.round(n[i_nr1]/k_r0)
        xm_r1 = xm[i_nr1]
        x_k0 = hist2array(xm_r0, n_r0)
        x_k1 = hist2array(xm_r1, n_r1)

        return x_k0, x_k1

    """
    Add valid min/max to plot (vertical lines + highlights)
    """
    def add_valid_minmax(self, fig, l, y, nz):

        vert_lines = []

        if not(self.valid_max is None):
          vert_lines.append(Span(location=self.valid_max, dimension='height', line_color='red', line_dash='dashed', 
            line_width=1.5))          
          vmax = np.where(l[nz] > self.valid_max)
          if len(vmax):
            fig.circle((l[nz])[vmax], (y[nz])[vmax], color='red', fill_color="white")

        if not(self.valid_min is None):
          vert_lines.append(Span(location=self.valid_min, dimension='height', line_color='red', line_dash='dashed', 
            line_width=1.5))          
          vmins = np.where(l[nz] < self.valid_min)
          if len(vmins):
            fig.circle((l[nz])[vmins], (y[nz])[vmins], color='red', fill_color="white")
        
        # Add vertical lines
        if vert_lines != []:
          fig.renderers.extend(vert_lines)

    """
    Boxplot (whiskers, rectangle, IQD)
    """
    def plot_boxplot(self, pw=1000, ph=450):

        # Read histograms into pandas
        df_stat = pd.read_csv(self.csv_stat)
        dates = get_dates_from_files(df_stat['File'])

        # generate some synthetic time series for six different categories
        cats = list("abcdef")
        yy = np.random.randn(2000)
        g = np.random.choice(cats, 2000)
        for i, l in enumerate(cats):
            yy[g == l] += i // 2
        df = pd.DataFrame(dict(score=yy, group=g))
        
        
        upper = df_stat['hist_q3'] + 1.5*df_stat['hist_IQD']
        lower = df_stat['hist_q1'] - 1.5*df_stat['hist_IQD']
        print(df_stat)
        # Figure config
        p = figure(tools="", width=pw, height=ph, background_fill_color="#efefef", x_axis_type='datetime')

        # stems
        p.segment(dates, upper, dates, df_stat['hist_q3'], line_color="black")
        p.segment(dates, lower, dates, df_stat['hist_q1'], line_color="black")

        # boxes (x,top,bott,width)
        p.vbar(dates, df_stat['hist_med'], df_stat['hist_q1'], 0.7, fill_color="#E08E79", line_color="black")
        p.vbar(dates, df_stat['hist_q3'], df_stat['hist_med'], 0.7, fill_color="#3B8686", line_color="black")

        # whiskers (almost-0 height rects simpler than segments)
        #p.rect(dates, lower.score, 0.2, 0.01, line_color="black")
        #p.rect(dates, upper.score, 0.2, 0.01, line_color="black")

        p.xgrid.grid_line_color = None
        p.ygrid.grid_line_color = "white"
        p.grid.grid_line_width = 2
        p.xaxis.major_label_text_font_size="12pt"

        output_file('/mnt/c/BSC/test.html')
        save(p)


    """
    Plot a single histogram, return figure
    """
    def plot_histogram(self, df, tag='total', col='File', pw=500, ph=250):

        # Get bins (left and right) + remove first element
        l = np.delete(df.columns.values, 0).astype(float) 
        xdist = abs(l[0]-l[1])
        r = [ l+xdist for v in l ]
        
        # Get count + remove first element
        y = np.delete(df.loc[df[col] == tag].values, 0).astype(float)

        # Get non-zero bins
        nz = np.nonzero(y)

        # Set up the figure same as before
        fig = figure(plot_width = pw, plot_height = ph,
                   title = 'Observed values histogram {}'.format(self.regions[tag]),
                   x_axis_label = self.xaxis_label, 
                   y_axis_label = 'Count')

        # Histograms and points where we have values
        fig.vbar(x=l[nz], width=xdist, bottom=0, top=y[nz], color="#941333")
        fig.dash(x=l[nz], y=y[nz], color="#941333")
        fig.toolbar.logo = None

        # Vertical lines indicating valid_min/valid_max if available
        self.add_valid_minmax(fig, l, y, nz)

        return fig

    """
    Plot multiple histograms into a grid and return a grid of the plots (file) 
    """
    def plot_histograms(self):

        # Read histograms into pandas
        df_hist = pd.read_csv(self.csv_hist)

        # Total histogram
        p0 = self.plot_histogram(df_hist, tag='total', pw=1000, ph=450)

        # Latitude regions
        p1 = None
        try:
          figs = []
          for t in [ 'N_polar', 'S_polar', 'N_midlat', 'S_midlat', 'N_hadley', 'S_hadley' ]:
            figs.append(self.plot_histogram(df_hist, tag=t))

          # Group plots into grid
          p1 = gridplot(figs, ncols=2, toolbar_options=dict(logo=None))
        except:
          pass # Dataset has no regions

        # Save to file        
        print('INFO: Saving plot in {}'.format(self.outfile_hist))
        output_file(self.outfile_hist)
        
        if p1 is None:
          save(p0)
        else:
          save(column(p0, p1))

hh="/data/dqc_test_data/reanalysis/reanalysis-era5-single-levels/sea_surface_temperature/ensemble_mean-hist.csv"
st="/data/dqc_test_data/reanalysis/reanalysis-era5-single-levels/sea_surface_temperature/ensemble_mean-stats.csv"
bp=bokehPlot(None,st,hh,None,None,None)
bp.plot_boxplot()